import numpy as np
import gdspy as gd
import gds_tools as gtools

#================================================
# Define healing function for transmission lines \\
#=========================================================================
# Initialization:   radius      :   radius of healing circle            ||
#                   endpoint    :   (x, y) of circle center             ||
#       (optional)  npoints     :   number of points to use for circum. ||
#       (optional)  layer       :   layer to put healer on              ||
#=========================================================================
def circle(radius, endpoint, npoints = 100, layer = 0, datatype = 0):

    c = gd.Round(endpoint, radius, number_of_points = npoints, layer = layer, datatype = datatype)
    c = gtools.classes.GDStructure(c, {'A': endpoint}, {'A': 2*radius})

    return c



import copy
import gdspy
import numpy as np
from operator import add, sub
from recordclass import recordclass

import gds_tools as gtools

class GDStructure:

    def __init__(self, structure, endpoints, endpoint_dims, endpoint_directions = None, method = None, args = {}):
        """Define class to store structures in

        Args:
            structure (gdspy.Path()): gdspy description of object (i.e. ouput of gdspy.PolyPath()).
            endpoints (dict): positions of the endpoints to which you connect other structures.
            endpoint_dims (dict): dimensions of the endpoints (width/height/etc.), to be used for connecting e.g. transmission lines.
            endpoint_directions ([type], optional): [description]. Defaults to None.
            method ([type], optional): [description]. Defaults to None.
            args (dict, optional): [description]. Defaults to {}.
        """

        self.type = 'GDStructure'
        self.structure = structure
        self.endpoints = endpoints
        self.endpoint_dims = endpoint_dims
        self.endpoint_directions = endpoint_directions
        self.compound = []
        self.prev = {}
        self.next = {}
        self.method = method

        self.__args_tuple__ = recordclass('args', args.keys())
        self.args = self.__args_tuple__(**args)

        self.gen = self.generate

    def generate(self):
        if self.method != None:
            return self.method(**self.args.__dict__)
        else:
            raise TypeError('This GDStructure does not support the .generate() method (yet)')

    def copy(self):
        """Makes a new copy of the structure,
        i.e. allocates new memory and copies the data contents to it.

        Returns:
        gds_tools.GDStructure: pointer of copy of original
        """

        # Create deepcopy of self and return the copy
        # Do not copy.deepcopy(self) directly, as this will also copy the connections and screw things up
        # Instead, just fill a new instance of the class with copies of relevant internal structure
        new_obj = gtools.classes.GDStructure(copy.deepcopy(self.structure), copy.deepcopy(self.endpoints), copy.deepcopy(self.endpoint_dims), copy.deepcopy(self.endpoint_directions))

        compound = []
        for i, c in enumerate(self.compound):
            c_copy = c.copy()
            compound.append(c_copy)
            new_obj.next[i] = c_copy

        new_obj.compound = compound

        return new_obj

    #==================
    # Rotate structure \\
    #=========================================================================
    # When rotating, all endpoints need to move wih the rotation,           ||
    # for this reason we need our own rotation function.                    ||
    #                                                                       ||
    # Arguments:    rad         :   amount of radians to rotate             ||
    #               signal_from :   tells linked structure who sent command ||
    #=========================================================================
    def rotate(self, rad, signal_from = None):

        if not isinstance(signal_from, list):
            signal_from = [signal_from]

        # Rotate the structure with standard gdspy .rotation() function
        if self not in signal_from:
            signal_from.append(self)
        self.structure.rotate(rad)

        # Rotate the endpoints
        for i in self.endpoints:
            self.endpoints[i] = tuple(gtools.funcs.VecRot(rad, self.endpoints[i]))
            if self.endpoint_directions:
                self.endpoint_directions[i] += rad

        # Rotate all connected structures
        for i in self.prev:
            if self.prev[i] not in signal_from and self.prev:
                signal_from.append(self.prev[i])
                self.prev[i].rotate(rad, signal_from = signal_from)

        for i in self.next:
            if self.next[i] not in signal_from and self.next:
                signal_from.append(self.next[i])
                self.next[i].rotate(rad, signal_from = signal_from)

        return self

    #=====================
    # Get structure layer \\
    #=========================================================================
    # Gdspy has some inconsistencies in how it stores layers, this getter   ||
    # function should make life easier to obtain the layer of the object    ||
    # stored in self.structure                                              ||
    #=========================================================================
    def getlayer(self):
        return self.structure.layers[0] if hasattr(self.structure, 'layers') else (self.structure.layer if hasattr(self.structure, 'layer') else 0)

    #=====================
    # Translate structure \\
    #=========================================================================
    # Arguments:    delta       :   vector (x, y) how much you want to move ||
    #               signal_from :   tells linked structure who sent command ||
    #=========================================================================
    def translate(self, delta, signal_from = None):

        if not isinstance(signal_from, list):
            signal_from = [signal_from]

        # Translate structure with standard gdspy .translate() function
        if self not in signal_from:
            signal_from.append(self)
        self.structure.translate(delta[0], delta[1])

        # Translate all the endpoints
        for i in self.endpoints:
            self.endpoints[i] = tuple(map(add, self.endpoints[i], delta))

        # Translate all connected structures
        for i in self.prev:
            if self.prev[i] not in signal_from and self.prev:
                signal_from.append(self.prev[i])
                self.prev[i].translate(delta, signal_from = signal_from)

        for i in self.next:
            if self.next[i] not in signal_from and self.next:
                signal_from.append(self.next[i])
                self.next[i].translate(delta, signal_from = signal_from)

        return self

    #=====================
    # Mirror structure \\
    #=========================================================================
    # Arguments:    p1, p2  :   p1 and p2 are (x, y) coords forming         ||
    #                           the mirror line                             ||
    #=========================================================================
    def mirror(self, p1, p2, signal_from = None):

        if not isinstance(signal_from, list):
            signal_from = [signal_from]

        # Mirror the gdspy shape
        if self not in signal_from:
            signal_from.append(self)
        self.structure.mirror(p1, p2 = p2)

        # Process the endpoints
        for k, v in self.endpoints.items():

            p1 = list(p1)
            p2 = list(p2)

            if p1[0] == p2[0]:
                p1[0] += 1E-20

            if p1[1] == p2[1]:
                p1[1] += 1E-20

            # y = ax + c : mirror line
            a = (p2[1] - p1[1]) / (p2[0] - p1[0])
            c = p1[1] - a * p1[0]
            d = (v[0] + (v[1] - c)*a) / (1 + a**2)

            v2x = 2*d - v[0]
            v2y = 2*d*a - v[1] + 2*c

            self.endpoints[k] = (v2x, v2y)

        # Ripple through all connected shapes
        for i in self.prev:
            if self.prev[i] not in signal_from and self.prev:
                signal_from.append(self.prev[i])
                self.prev[i].mirror(p1, p2, signal_from = signal_from)

        for i in self.next:
            if self.next[i] not in signal_from and self.next:
                signal_from.append(self.next[i])
                self.next[i].mirror(p1, p2, signal_from = signal_from)

        return self

    #=================
    # Heal connection \\
    #=========================================================================
    # When connection endpoints, there is always a little space that is not ||
    # overlapped. This function will fill this overlap.                     ||
    #                                                                       ||
    # Arguments:    endpoint    :   (x, y) center of healer                 ||
    #    (optional) npoints     :   number of points to use for boundary    ||
    #=========================================================================
    def heal(self, endpoint, npoints = 100, r = 'auto', layer = None, datatype = None):

        if type(r) == str and r == 'auto':
            r = self.endpoint_dims[endpoint] / 2

        healer = gtools.heal.circle(r, self.endpoints[endpoint], npoints = npoints, layer = layer if layer != None else self.getlayer(), datatype = datatype if datatype != None else self.structure.datatypes[0])

        self.prev['HEAL_' + endpoint] = healer
        self.compound += [healer]

        return self

    #===================
    # Connect structure \\
    #=========================================================================
    # Arguments:    ep_self :   reference to A, B, C, etc.                  ||
    #                           name of endpoint of this structure          ||
    #               to      :   GDStructure class input to which to connect ||
    #               ep_to   :   endpoint to which to connect to             ||
    #               offset  :   (x, y) offset from target endpoints         ||
    #               obj_link:   use Python object reference for linked list ||
    #                           entry rather than endpoint key              ||
    #=========================================================================
    def connect(self, ep_self, to, ep_to, offset = (0, 0), obj_link = False, rotate = True):

        # Rotate to correct orientation
        if self.endpoint_directions != None and to.endpoint_directions != None and self.endpoint_directions and to.endpoint_directions and rotate:
            self.rotate((to.endpoint_directions[ep_to] - self.endpoint_directions[ep_self] - np.pi) % (2 * np.pi))

        # Move to connect endpoints
        delta = tuple(map(sub, to.endpoints[ep_to], self.endpoints[ep_self]))
        delta = tuple(map(add, delta, offset))
        self.mov(delta)

        # Update linked list
        if not obj_link:
            to.next[ep_to] = self
            self.prev[ep_self] = to
        else:
            to.next[self] = self
            self.prev[to] = to

        return self

    #======================
    # Disconnect structure \\
    #=========================================================================
    # Only removes the references in the linked lists                       ||
    #=========================================================================
    def disconnect(self):

        for e in self.endpoints:
            if e in self.next:
                for ne in self.next[e].endpoints:
                    if ne in self.next[e].prev and self.next[e].prev[ne] == self:
                        del self.next[e].prev[ne]
                del self.next[e]
            if e in self.prev:
                for pe in self.prev[e].endpoints:
                    if pe in self.prev[e].next and self.prev[e].next[pe] == self:
                        del self.prev[e].next[pe]
                del self.prev[e]

        return self

    # Aliases
    rot = rotate
    mov = translate
    con = connect
    dis = disconnect



import numpy as np
import gdspy as gd
import gds_tools as gtools

#=======================
# Generate a hollow box \\
#=========================================================================
# Arguments:    size            :   (x, y) specifiying box outer size   ||
#               border_width    :   width of the border                 ||
#=========================================================================
def hollow_box(size, border_width, layer = 0):

    outer = gd.Rectangle((0, 0), size)
    inner = gd.Rectangle((border_width, border_width), (size[0] - border_width, size[1] - border_width))

    hbox = gd.fast_boolean(outer, inner, 'not', layer = layer)
    ends = {'A': (0, size[1] / 2), 'B': (size[0] / 2, 0), 'C': (size[0] / 2, size[1]), 'D': (size[0], size[1] / 2), 'CENTER': (size[0] / 2, size[1] / 2)}
    epsz = {'A': size[1], 'CENTER': None}

    return gtools.classes.GDStructure(hbox, ends, epsz)

#========================
# Generate a regular box \\
#=========================================================================
# Arguments:    size            :   (x, y) specifiying box outer size   ||
#=========================================================================
def box(size, layer = 0, datatype = 0):

    box = gd.Rectangle((0, 0), size, layer = layer, datatype = datatype)
    ends = {'A': (0, size[1] / 2), 'B': (size[0] / 2, 0), 'C': (size[0] / 2, size[1]), 'D': (size[0], size[1] / 2), 'CENTER': (size[0] / 2, size[1] / 2),
            'W': (0, 0), 'X': (size[0], 0), 'Y': (size[0], size[1]), 'Z': (0, size[1])}
    epsz = {'A': size[1], 'D': size[1], 'B': size[0], 'C': size[0], 'CENTER': None, 'W': size[0], 'X': size[0], 'Y': size[0], 'Z': size[0]}

    return gtools.classes.GDStructure(box, ends, epsz)

#===================
# Generate a circle \\
#=========================================================================
# Arguments:    r           :   radius                                  ||
#    (optional) resolution  :   number of points in [0..2pi]            ||
#=========================================================================
def circle(r, resolution = 100, layer = 0):

    circle = gd.Round((0, 0), r, number_of_points = resolution, layer = layer)
    ends = {'CENTER': (0, 0), 'BOTTOM': (0, -r)}
    epsz = {'CENTER': r, 'BOTTOM': None}

    return gtools.classes.GDStructure(circle, ends, epsz)

def triangle(side_1, side_2, angle, layer = 0):

    p = [(0, 0), (side_1, 0), gtools.funcs.VecRot(angle, (side_2, 0))]

    poly = gd.Polygon(p, layer = layer)
    ends = {gtools.alphabet[i]: p[i] for i in range(0, len(p))}
    epsz = {gtools.alphabet[i]: None for i in range(0, len(p))}

    return gtools.classes.GDStructure(poly, ends, epsz)




import sys,os
import copy
import gdspy as gd
import numpy as np
import gds_tools as gtools
import cProfile, pstats, io
from recordclass import recordclass

def profile(fnc):

    """A decorator that uses cProfile to profile a function"""

    def inner(*args, **kwargs):

        pr = cProfile.Profile()
        pr.enable()
        retval = fnc(*args, **kwargs)
        pr.disable()
        s = io.StringIO()
        sortby = 'cumulative'
        ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
        ps.print_stats()
        print(s.getvalue())
        return retval

    return inner

def RotMat(rad):
    #==========================
    # Generate rotation matrix \\
    #=========================================================================
    # Arguments:    rad     :   radians to rotate about origin              ||
    #=========================================================================
    return np.matrix([[np.cos(rad), -np.sin(rad)], [np.sin(rad), np.cos(rad)]])

def VecRot(rad, vec, origin = (0, 0)):
    #=========================
    # Perform vector rotation \\
    #=========================================================================
    # Arguments:    rad     :   radians to rotate about origin              ||
    #               vec     :   input vector (2-x-n list)                   ||
    #=========================================================================
    return (RotMat(rad).dot(np.array(vec) - np.array(origin)) + np.array(origin)).tolist()[0]

def instruction_parse(s, args = None):
    #============================
    # Simple instructions parser \\
    #=========================================================================
    # Parses a string and converts it to a dictionary.                      ||
    #                                                                       ||
    # Arguments:    s       :   input strung                                ||
    #               args    :   dictionary with keys for variable placement ||
    #=========================================================================
    if args:
        for a in args:
            s = s.replace('{'+a+'}', str(args[a]))

    dic = {}
    key = ''
    val = ''
    pos = 'key'
    for i, c in enumerate(s):
        if c == ' ' or c == '\n' or c == '\t': # ignore whitespace
            continue
        elif c == ':':
            pos = 'val'
        elif c != ':' and c != ',' and pos == 'key':
            key += c
        elif c != ':' and c != ',' and pos == 'val':
            val += c
        elif c == ',':
            True # do nothing
        else:
            print('Error: unknown parameter, could not parse.')
            return False

        if c == ',' or (i + 1) == len(s):
            val = eval(val.replace('pi', str(np.pi)))
            dic[key] = float(val)
            key = ''
            val = ''
            pos = 'key'
            if (i + 1) == len(s):
                break

    return dic

def flatten(objectlist, endpoints, endpoint_dims, layer = 0):
    #===========================
    # FLatten a list of objects \\
    #=========================================================================
    # Flattening will cause all objects in the objectlist to be placed in   ||
    # one single layer and remove boundaries between them if there are any. ||
    # All layer information will become lost! If you just want to combine   ||
    # structures while keeping layer information, use cluster()             ||
    #                                                                       ||
    # Arguments:    objectlist      :   list of objects (GDStructure)       ||
    #               endpoints       :   dictionary of new endpoints         ||
    #               endpoint_dims   :   dictionary of new endpoint sizes    ||
    #=========================================================================
    # Define function to allow for recursive walk through list and pick out all
    # compound structures
    def stacker(inlist):

        outlist = []
        for i in inlist:
            if i.compound:
                outlist += [i] + stacker(i.compound)
            else:
                outlist += [i]

        return outlist

    objectlist = stacker(objectlist)

    ends = copy.deepcopy(endpoints)
    epsz = copy.deepcopy(endpoint_dims)

    objs = []
    for i in objectlist:
        objs.append(i.structure)

    return gtools.classes.GDStructure(gd.fast_boolean(objs, None, 'or', layer = layer), ends, epsz)

def lattice(cell, repeat, spacing):
    #============================
    # Generate a crystal lattice \\
    #=========================================================================
    # Arguments:    cell        :   unit cell as gdspy Cell object          ||
    #               repeat      :   (n_x, n_y) vector with amount of cells  ||
    #               spacing     :   sapce between unit cells                ||
    #=========================================================================
    array = gd.CellArray(cell, repeat[0], repeat[1], spacing)
    ends = {'A': (0, spacing[1] * repeat[1] / 2), 'B': (spacing[0] * (repeat[0] - 1) / 2, spacing[1] * (repeat[1] - 1/2))}
    epsz = {'A': 0, 'B': 0}

    return gtools.classes.GDStructure(array, ends, epsz)

def lattice_cutter(lattice, objectlist, mode = 'and', layer = 0):
    #=====================================
    # Cut a lattice up using fast_boolean \\
    #=========================================================================
    # Arguments:    lattice     :   output of lattice() function            ||
    #               objectlist  :   list of objects that intersect lattice  ||
    #    (optional) mode        :   what boolean operation to apply         ||
    #    (optional) layer       :   layer to put resulting structure on     ||
    #=========================================================================
    if type(objectlist) is not type([]):
        objectlist = [objectlist]

    for i in objectlist:
        if i.compound:
            lattice = lattice_cutter(lattice, i.compound)
        lattice.structure = gd.fast_boolean(lattice.structure, i.structure, mode, layer = layer)

    return lattice

def cluster(objectlist, endpoints, endpoint_dims, endpoint_directions = None, method = None, args = {}, ignore_conflict = False):
    #============================
    # Cluster list of structures \\
    #=========================================================================
    # Will combine GDStructure objects in a list so they will act as one    ||
    # single structure. They will keep their individual properties.         ||
    #                                                                       ||
    # Arguments:    objectlist      :   list of GDStructure objects         ||
    #               endpoints       :   dict of endpoints                   ||
    #               endpoint_dims   :   dict of endpoint sizes              ||
    #=========================================================================
    if not ignore_conflict:
        for i in endpoints:
            if i in objectlist[0].endpoints:
                print('Error: this endpoint already exists in first element of given list of structures to combine.\nCannot continue, as overwriting existing endpoints will result in broken compound structure.\nTo fix this, give different names to your new endpoints as input to this functions.\n')
                return False

    objectlist[0].endpoints = endpoints
    objectlist[0].endpoint_dims = endpoint_dims
    if endpoint_directions != None:
        objectlist[0].endpoint_directions = endpoint_directions
    else:
        objectlist[0].endpoint_directions = {}
    objectlist[0].compound += objectlist[1:]

    objectlist[0].method = method
    objectlist[0].__args_tuple__ = recordclass('args', args.keys())
    objectlist[0].args = objectlist[0].__args_tuple__(**args)

    for i, object in enumerate(objectlist[1:]):
        objectlist[0].next['COMPOUND_'+str(i)] = object
        object.prev['COMPOUND_0'] = objectlist[0]

    return objectlist[0]

def add(cell, objectlist):
    #================================
    # Add structures to a gdspy cell \\
    #=========================================================================
    # Arguments:    cell        :   gdspy cell object                       ||
    #               objectlist  :   list of GDStructure objects             ||
    #=========================================================================
    if type(objectlist) is not type([]):
        objectlist = [objectlist]

    structs = []
    for i in objectlist:
        if i.structure:
            if type(i.structure) is type([]):
                for structure in i.structure:
                    structs.append(structure)
            else:
                structs.append(i.structure)

            if i.compound:
                gtools.add(cell, i.compound)

    cell.add(structs)

    return cell

def mirror(p):
    #============================
    # Mirror points about y-axis \\
    #=========================================================================
    # Arguments:    p   :   list of (x, y) points                           ||
    #=========================================================================
    for i, val in enumerate(p):
        p[i] = (-val[0], val[1])

    return p

def symm_coords(points, mirror_x = True, mirror_y = True):

    if type(points) is not 'list':
        points = [points]

    output_points = copy.deepcopy(points)

    for i, val in enumerate(points):
        if mirror_y:
            output_points.append((-val[0], val[1]))
        if mirror_x:
            output_points.append((val[0], -val[1]))
        if mirror_x and mirror_y:
            output_points.append((-val[0], -val[1]))
    return output_points

def save(cell, filename, unit = 1e-6, precision = 1e-9):
    #=====================
    # Save cell to a file \\
    #=========================================================================
    # Arguments:    cell        :   gdspy cell object or a list of cells    ||
    #               filename    :   filename to write to (relative path)    ||
    #=========================================================================
    writer = gd.GdsWriter(filename, unit = unit, precision = precision)
    print(cell)
    if type(cell) == type([]):
        for cell_item in cell:
            writer.write_cell(cell_item)
    else:
        writer.write_cell(cell)

    return writer.close()

def inside(points, cellref, dist, nop = 3, precision = 0.001):
    #=====================
    # Save cell to a file \\
    #=========================================================================
    # Arguments:    points      :   list of points to check                 ||
    #               cellref     :   gdspy cell reference object             ||
    #               dist        :   distance from points to search          ||
    #               nop         :   number of probe points within dist      ||
    #               precision   :   gdspy.inside precision parameter        ||
    #=========================================================================
    # Force uneven
    if nop % 2 == 0:
        nop += 1

    search_ps = []
    for p in points:
        px = np.linspace(p[0] - dist/2, p[0] + dist/2, nop)
        py = np.linspace(p[1] - dist/2, p[1] + dist/2, nop)
        search_ps.append([[i, j] for i in px for j in py])

    return gd.inside(search_ps, cellref, precision = precision)



import numpy as np
import scipy.spatial as sp
import copy
import gdspy as gd
import gds_tools as gtools

#===========================
# Make a transmission line  \\
#=========================================================================
# Arguments:    dxy         :   (dx, dy), lengths of the line           ||
#               width       :   width of the line                       ||
#    (optional) width_end   :   different width at the end              ||
#    (optional) layer       :   choose layer to put spiral onto         ||
#=========================================================================
def line(dxy, width, width_end = False, layer = 0, datatype = 0):

    width_end = width if (not width_end) else width_end

    p = [(0, 0), dxy]
    w = [width, width_end]

    poly = gd.PolyPath(p, w, layer = layer, datatype = datatype)
    ends = {'A': p[0], 'B': p[-1]}
    epsz = {'A': w[0], 'B': w[-1]}

    return gtools.classes.GDStructure(poly, ends, epsz)

#================================================
# Generate a meander \\ Author: Marta Pita Vidal \\
#=========================================================================
# Arguments:    n               :   number of meanders                  ||
#               width           :   width of the meander path           ||
#               radius          :   radius of the corner circles        ||
#               meander_length  :   length from corner to corner        ||
#               connect_length  :   length of initial transmission line ||
#    (optional) layer           :   layer to put meander on             ||
#=========================================================================
def meander(n, width, radius, meander_length, connect_length, layer = 0):

    # Aliases
    w = width
    d = radius
    N = n
    position = (0, 0)
    xMeander = meander_length / 2
    yBase = connect_length

    # Start path
    meander_path = gd.Path(w, (0, 0))
    meander_path.segment(yBase, np.pi/2, layer = layer).turn(w/2, 'r', layer = layer).segment(xMeander - w/2, layer = layer).turn(d, 'll', layer = layer).segment(xMeander, layer = layer)

    # Generate periodic meanders
    for i in range(N - 1):
        meander_path.segment(xMeander, layer = layer).turn(d, 'rr', layer = layer).segment(2 * xMeander, layer = layer).turn(d, 'll', layer = layer).segment(xMeander, layer = layer)

    # Finalize
    meander_path.segment(xMeander, layer = layer).turn(d, 'rr', layer = layer).segment(xMeander - w/2, layer = layer).turn(w/2, 'l', layer = layer).segment(yBase, '+y', layer = layer)

    # Create object
    struct = gtools.classes.GDStructure(meander_path, {'A': (0, 0), 'B': (meander_path.x, meander_path.y)}, {'A': w, 'B': w})

    return struct

#===================================
# Make a transmission line splitter \\
#=========================================================================
# Arguments:    n       :   number of splits                            ||
#               width   :   width of the paths                          ||
#               length  :   total length from start to end              ||
#               space   :   spacing between splits                      ||
#    (optional) layer   :   choose layer to put split onto              ||
#=========================================================================
def splitter(n, width, length, space, layer = 0):

    p = [(0, 0), (length / 2 - width / 2, 0), (length / 2 - width / 2, (n*width + (n-1)*space)/2 - width / 2)]

    ends = {'A': (0, -width / 2)}
    epsz = {'A': width}

    for i in range(0, n):
        p += [(length, (n*width + (n-1)*space)/2 - i*(space + width) - width / 2)]

        ends[gtools.alphabet[i + 1]] = (length, p[-1][1] - width / 2)
        epsz[gtools.alphabet[i + 1]] = width

        p += [(length, (n*width + (n-1)*space)/2 - i*(space + width) - width - width / 2)]
        p += [(length / 2 + width / 2, (n*width + (n-1)*space)/2 - i*(space + width) - width - width / 2)]
        p += [(length / 2 + width / 2, (n*width + (n-1)*space)/2 - i*(space + width) - width - space - width / 2)]

    p = p[:-1]

    p += [(length / 2 - width / 2, (n*width + (n-1)*space)/2 - (n-1)*(space + width) - width - width / 2)]
    p += [(length / 2 - width / 2, -width)]
    p += [(0, -width)]

    poly = gd.Polygon(p, layer = layer)

    return gtools.classes.GDStructure(poly, ends, epsz)

# Helper function, returns indeces of neighbouring points (u, d, l, r) in a grid
def lookaround(index, row_len, pref = 'y'):

    n = []
    n.append(index - row_len)
    n.append(index + row_len)
    if index % row_len != 0:
        n.append(index - 1)
    n.append(index + 1)

    if pref == 'x':
        n = n[::-1]

    return n

# Autoroute using Lee's routing algorithm
def router(cell, fr_str, fr_ep, to_str, to_ep, width, bmul, grid_s = 1, xr = False, yr = False, uniform_width = False, precision = 0.001, pref = 'y', switch_pref = False, layer = 0, debug = False, nop = 21, dist_multi = 2, pathmethod = 'poly', detect = 'native'):

    fr = fr_str.endpoints[fr_ep]
    to = to_str.endpoints[to_ep]

    border_s = bmul * grid_s
    box = [fr, to]

    # Make sure first box coord is always the top-left corner and add additional border points
    xs = [box[0][0], box[1][0]]
    ys = [box[0][1], box[1][1]]
    box = [[min(xs) - border_s, max(ys) + border_s], [max(xs) + border_s, min(ys) - border_s]]

    # Build list of gridpoints that are outside all structures in cell
    lxr = int((box[1][0] - box[0][0]) / grid_s) + 1
    lyr = int((box[0][1] - box[1][1]) / grid_s) + 1

    if type(xr) not in [list, np.ndarray]:
        xr = np.linspace(box[0][0], box[1][0], lxr)
    else:
        lxr = len(xr)

    if type(yr) not in [list, np.ndarray]:
        yr = np.linspace(box[0][1], box[1][1], lyr)
    else:
        lyr = len(yr)

    p = []
    p_d_to = []
    p_d_fr = []
    for y in yr:
        for x in xr:

            c = (x, y)
            p.append(c)

            # Compute squared Euclidean distance from to and fr coords for each gridpoint
            # For optimization we don't need to sqrt() since minimal in squared is minimal in sqrt
            dist_fr = (x - fr[0])**2 + (y - fr[1])**2
            dist_to = (x - to[0])**2 + (y - to[1])**2

            p_d_fr.append(dist_fr)
            p_d_to.append(dist_to)

    p_i = np.array(p)
    p_d_fr = np.array(p_d_fr)
    p_d_to = np.array(p_d_to)

    # Build list of points that are inside a structure
    cell_ref = gd.CellReference(cell)
    if detect == 'native':
        inside = np.array(gd.inside(p, cell_ref, precision = precision))
    elif detect == 'custom':
        inside = np.array(gtools.funcs.inside(p, cell_ref, dist = dist_multi*grid_s, nop = nop, precision = precision))
    else:
        raise ValueError('Parameter \'detect\' is only allowed to have values [\'native\', \'custom\'], cannot continue')

    p_d_fr_min = np.min(p_d_fr[np.argwhere(inside == False)])
    p_d_to_min = np.min(p_d_to[np.argwhere(inside == False)])

    # Get p_i index of starting values
    start_i = np.argwhere(p_d_fr == p_d_fr_min).tolist()
    end_i = np.argwhere(p_d_to == p_d_to_min).tolist()

    start_i = [item for sublist in start_i for item in sublist]
    end_i = [item for sublist in end_i for item in sublist]

    start = p_i[start_i]
    end = p_i[end_i]

    # Now start stepping from start to end, labelling all gridpoints accordingly by the number of steps required from starting point to reach it
    n = [0] * 4
    lp = len(p)
    p_g = [0] * lp
    path_found = False
    k = 0

    while not path_found and start_i:

        k += 1
        next_start_i = []

        if debug:
            print(start_i)

        for i in start_i:

            # Look up, down, left and right, store the index
            n = lookaround(i, lxr, pref = pref)

            # Check if any of the neighbouring points are not in a structure and not in p_g
            for nb in n:

                if nb in end_i:
                    path_found = True
                    p_g[nb] = k
                    final_index = nb

                    if debug:
                        # Visualize
                        circ = gd.Round(p[nb], 0.1, layer = 10)
                        cell.add(circ)
                        txt = gd.Text(str(k), 0.5, p[nb], layer = 11)
                        cell.add(txt)

                    break

		         # Point is out of bounds, marked as structure (< 0) or already has a step value (> 0)
                if nb < 0 or nb >= lp or p_g[nb] != 0 or (i % lxr == 0 and nb % lxr == 1) or (i % lxr == 1 and nb % lxr == 0):
                    continue # Skip this iteration

                if inside[nb]:
                    p_g[nb] = -1
                else:
                    p_g[nb] = k
                    next_start_i.append(nb)

                    if debug:
                        # Visualize
                        circ = gd.Round(p[nb], 0.1, layer = 1)
                        cell.add(circ)
                        txt = gd.Text(str(k), 0.5, p[nb], layer = 2)
                        cell.add(txt)

        start_i = copy.copy(next_start_i)

    # Routing ended, checking whether we succeeded
    if not path_found:
        print('>> ERROR: No existing route was found.')
        return False

    print('>> Found a route in ' + str(k) + ' steps.')

    # Backtrace path
    this_index = final_index
    backtraced = [to, p[final_index]]
    switched = False
    for this_k in range(k, -1, -1):

        # Change move preference after switch_pref moves
        if switch_pref and not switched and this_k < switch_pref*k:
            pref = 'x' if pref == 'y' else 'y'
            switched = True

        n = lookaround(this_index, lxr, pref = pref)
        for nb in n:
            if nb < lp and p_g[nb] == this_k:
                this_index = nb
                backtraced.append(p[nb])
                break

    backtraced.append(fr)

    if debug:
        print('>> Points of found route:')
        print(backtraced)

    # Generate list of widths for the route
    if not uniform_width:
        to_w = to_str.endpoint_dims[to_ep]
        fr_w = fr_str.endpoint_dims[fr_ep]
        ws = [to_w if to_w != None else width]*2 + [width]*(len(backtraced)-4) + [fr_w if fr_w != None else width]*2
    else:
        ws = width

    # Create backtraced path
    if pathmethod == 'poly':
        r = gd.PolyPath(backtraced, ws, layer = layer)
    elif pathmethod == 'flex':
        r = gd.FlexPath(backtraced, ws, corners = 'smooth', layer = layer)
    else:
        raise ValueError('Parameter \'pathmethod\' only has allowed values [\'poly\', \'flex\']')

    ends = {'A': backtraced[0], 'B': backtraced[-1]}
    epsz = {'A': ws[0] if not uniform_width else width, 'B': ws[-1] if not uniform_width else width}

    structure = gtools.classes.GDStructure(r, ends, epsz)

    # Connect to 'to' and 'from' structures
    fr_str.next['AUTOROUTE_A'] = structure
    to_str.next['AUTOROUTE_B'] = structure
    structure.prev['AUTOROUTE_A'] = fr_str
    structure.prev['AUTOROUTE_B'] = to_str

    return structure



# Module definitions
import copy
from gds_tools import functions, classes, routing, heal, geometry

# Handy constants
a = 'A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z'.split(',')
alphabet = copy.copy(a)
for i in range(1, 20):
    for c in a:
        alphabet.append(c + str(i))

# Aliases
GDStructure = classes.GDStructure
struct = classes.GDStructure
instruction_parse = functions.instruction_parse
add = functions.add
save = functions.save
flatten = functions.flatten
cluster = functions.cluster
lattice = functions.lattice
lattice_cutter = functions.lattice_cutter

# Backwards compatibility
funcs = functions
transmission = routing



from setuptools import setup, find_packages

setup(
    name='gds_tools',
    version='0.1.1',

    install_requires=[
        'numpy',
        'gdspy',
        'recordclass'
    ],

    license = 'MIT',

    author='Lieuwe Stek',
    author_email='lieuwe.stek@gmail.com',

    description=("Package aiding scripted design of structures for nanofabrication of devices for condensed matter physics research."),

    packages=find_packages(),

    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.6'
        ],

    keywords='gds gdsii cad script scripting physics nanofabrication condensed matter',

    url='https://github.com/kouwenhovenlab/gds_tools'
)
